// gluggMoreShapes
// Creates a few standard shapes not covered in gluggShapes:
// Sphere, Cone, Cylinder and Ellipsoid
// By Ingemar 2021.

#include <stdlib.h>
#ifdef __APPLE__
	#include <OpenGL/gl3.h>
// uses framework Cocoa
#else
	#include <GL/gl.h>
#endif
#include "MicroGlut.h"
#include "GL_utilities.h"
#include "VectorUtils3.h"
#include "glugg.h"
#include "LittleOBJLoader.h"

#define min(a,b) (((a)<(b))?(a):(b))
#define max(a,b) (((a)>(b))?(a):(b))

GLuint program; // Shader
mat4 projectionMatrix;

// Make ellipsoid, using the shape of a sphere and resize.
// Note that I use 1/width, 1/height etc for normals!
vec3 *MakeEllipsoid(int *count, int aSlices, int hSlices, vec3 dimensions)
{
	float height = dimensions.y;
	float width = dimensions.x;
	float depth = dimensions.z;
	
	vec3 top = SetVector(0,height,0);
	vec3 bottom = SetVector(0,-height,0);
	vec3 bn = SetVector(0,-1,0); // Bottom normal
	vec3 tn = SetVector(0,1,0); // Top normal

	gluggBegin(GLUGG_TRIANGLE_STRIP);
// Create bands
	for (float b = -M_PI/2 + M_PI / hSlices; b < M_PI/2 - M_PI / hSlices + 0.0001; b += M_PI / hSlices)
	{
		gluggMode(GLUGG_TRIANGLE_STRIP);
		for (float a = 0.0; a < 2.0*M_PI+0.0001; a += 2.0*M_PI / aSlices)
		{
			float y1 = sin(b - M_PI / hSlices);
			float y2 = sin(b);
			float a0 = a - 2.0*M_PI / aSlices;
			float a1 = a;
			
			float width1 = sqrt(1 - y1 * y1) * width;
			float width2 = sqrt(1 - y2 * y2) * width;
			
			vec3 p1 = SetVector(width1 * cos(a1), y1, width1 * sin(a1));
			vec3 p2 = SetVector(width2 * cos(a1), y2, width2 * sin(a1));
			
			p1 = SetVector(p1.x*width, p1.y*height, p1.z*depth);
			p2 = SetVector(p2.x*width, p2.y*height, p2.z*depth);

			vec3 p1n = normalize(SetVector(p1.x/width, p1.y/height, p1.z/depth));
			vec3 p2n = normalize(SetVector(p2.x/width, p2.y/height, p2.z/depth));

	// Done making points and normals. Now create polygons!
			gluggNormalv(p2n);
		    gluggVertexv(p2);
			gluggNormalv(p1n);
		    gluggVertexv(p1);
		}
	}

//	// Walk around top
	gluggMode(GLUGG_TRIANGLE_FAN); // Reset to new fan
	float y = sin(M_PI/2 - M_PI / hSlices);
	float width1 = sqrt(1 - y * y);
	gluggNormalv(tn);
	gluggVertexv(top);
	for (float a = 0.0; a < 2.0*M_PI+0.0001; a += 2.0*M_PI / aSlices)
	{
		vec3 p = SetVector(width1 * cos(a), y, width1 * sin(a));

		p = SetVector(p.x*width, p.y*height, p.z*depth);
		vec3 pn = normalize(SetVector(p.x/width, p.y/height, p.z/depth));
		
		gluggNormalv(pn);
	    gluggVertexv(p);
	}
//	// Walk around bottom
	gluggMode(GLUGG_TRIANGLE_FAN); // Reset to new fan
	y = sin(-M_PI/2 + M_PI / hSlices);
	gluggNormalv(bn);
	gluggVertexv(bottom);
	for (float a = 2.0*M_PI; a >= -0.001; a -= 2.0*M_PI / aSlices)
	{
		vec3 p = SetVector(width1 * cos(a), y, width1 * sin(a));

		p = SetVector(p.x*width, p.y*height, p.z*depth);
		vec3 pn = normalize(SetVector(p.x/width, p.y/height, p.z/depth));
		
		gluggNormalv(pn);
	    gluggVertexv(p);
	}

	// Then walk around the bottom
	return gluggEnd(count, program, 0);
}

// Make a sphere using the ellipsoid
vec3 *MakeSphere(int *count, int aSlices, int hSlices, float radius)
{
	return MakeEllipsoid(count, aSlices, hSlices, SetVector(radius, radius, radius));
}

vec3 *MakeCylinder(int *count, int aSlices, float height, float width)
{
	gluggBegin(GLUGG_TRIANGLE_STRIP);

	vec3 top = SetVector(0,height,0);
	vec3 center = SetVector(0,0,0);
	vec3 bn = SetVector(0,-1,0); // Bottom normal
	vec3 tn = SetVector(0,1,0); // Top normal
	
	for (float a = 0.0; a < 2.0*M_PI+0.0001; a += 2.0*M_PI / aSlices)
	{
		float a1 = a;
		
		vec3 p1 = SetVector(width * cos(a1), height, width * sin(a1));
		vec3 p2 = SetVector(width * cos(a1), 0, width * sin(a1));
		vec3 pn = SetVector(cos(a1), 0, sin(a1));

// Done making points and normals. Now create polygons!
		gluggNormalv(pn);
	    gluggVertexv(p2);
	    gluggVertexv(p1);
	}
	
	// Then walk around the top and bottom with fans
	gluggMode(GLUGG_TRIANGLE_FAN);
	gluggNormalv(bn);
	gluggVertexv(center);
	// Walk around edge
	for (float a = 0.0; a <= 2.0*M_PI+0.001; a += 2.0*M_PI / aSlices)
	{
		vec3 p = SetVector(width * cos(a), 0, width * sin(a));
	    gluggVertexv(p);
	}
	// Walk around edge
	gluggMode(GLUGG_TRIANGLE_FAN); // Reset to new fan
	gluggNormalv(tn);
	gluggVertexv(top);
	for (float a = 2.0*M_PI; a >= -0.001; a -= 2.0*M_PI / aSlices)
	{
		vec3 p = SetVector(width * cos(a), height, width * sin(a));
	    gluggVertexv(p);
	}

	return gluggEnd(count, program, 0);
}

vec3 *MakeCone(int *count, int aSlices, float height, float width)
{
	gluggBegin(GLUGG_TRIANGLES);
	// GLUGG_TRIANGLE_FAN not suitable for sides since the top vertex needs different normals for different polygons.

	vec3 top = SetVector(0,height,0);
	vec3 center = SetVector(0,0,0);
	vec3 bn = SetVector(0,-1,0); // Bottom normal
	
	for (float a = 0.0; a < 2.0*M_PI; a += 2.0*M_PI / aSlices)
	{
		// Make two triangles, one to center, one to top (0,height,0).
		
		float a0 = a - 2.0*M_PI / aSlices;
		float a1 = a;
		
		vec3 p = SetVector(width * cos(a1), 0, width * sin(a1));
		vec3 q = SetVector(width * cos(a0), 0, width * sin(a0));

		vec3 n = normalize(SetVector(width, height, 0));
		vec3 pn = SetVector(n.x * cos(a1), n.y, n.x * sin(a1));
		vec3 qn = SetVector(n.x * cos(a0), n.y, n.x * sin(a0));

// Done making points and normals. Now create polygons!
		gluggNormalv(pn);
	    gluggVertexv(p);
		gluggNormalv(normalize(VectorAdd(pn, qn)));
	    gluggVertexv(top);
	    gluggVertexv(q);
	    
		gluggNormalv(bn);
	    gluggVertexv(p);
	    gluggVertexv(q);
	    gluggVertexv(center);
	}
	
	// Then walk around the bottom with a fan
	gluggMode(GLUGG_TRIANGLE_FAN);	
	gluggNormalv(bn);
	gluggVertexv(center);
	// Walk around edge
	for (float a = 0.0; a < 2.0*M_PI; a += 2.0*M_PI / aSlices)
	{
		vec3 p = SetVector(width * cos(a), 0, width * sin(a));
	    gluggVertexv(p);
	}
	
	return gluggEnd(count, program, 0);
}


static GLfloat red[4] = { 0.8, 0.1, 0.0, 1.0 };
static GLfloat green[4] = { 0.0, 0.8, 0.2, 1.0 };
static GLfloat blue[4] = { 0.2, 0.2, 1.0, 1.0 };
static GLfloat yellow[4] = { 0.8, 0.8, 0.0, 1.0 };
static GLfloat magenta[4] = { 1.0, 0.2, 1.0, 1.0 };
static GLfloat cyan[4] = { 0, 1, 1, 1.0 };

GLuint s1, s2, s3, s4;
int count1, count2, count3, count4;
GLuint s0;
int count0;

void display(void)
{
	float time = glutGet(GLUT_ELAPSED_TIME) / 1000.0;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    mat4 cam = lookAt(0, 2, 8, 0, 0, 0, 0,1,0);
	
	mat4 m;
    m = Mult(cam, Mult(Mult(T(2.0, 2.0,0.0), Ry(time*2)), Rx(3*sin(time/5))));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, red);
	glBindVertexArray(s1);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, count1);

    m = Mult(cam, Mult(Mult(T(-2.0, -2.0,0.0), Ry(time*2)), Rx(5*sin(time/2))));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, blue);
	glBindVertexArray(s2);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, count2);

    m = Mult(cam, Mult(Mult(T(-2.0, 1.4,0.0), Ry(time*2)), Rx(2*sin(time/5))));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, green);
	glBindVertexArray(s3);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, count3);

    m = Mult(cam, Mult(Mult(T(2.0, -2.0,0.0), Ry(time*2)), Rx(3*sin(time/5))));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, yellow);
	glBindVertexArray(s4);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, count4);

    glutSwapBuffers();
}

void myinit()
{
    glEnable(GL_DEPTH_TEST); /* enable z buffer */
    glDisable(GL_CULL_FACE);

	program = loadShaders("sphere.vert", "sphere.frag");

	s1 = MakeSphere(&count1, 15, 10, 1.0);
	s2 = MakeCone(&count2, 20, 2.0, 1.0);
	s3 = MakeCylinder(&count3, 20, 1.5, 1.0);
	s4 = MakeEllipsoid(&count4, 15, 14, SetVector(1, 2, 1.5));

	projectionMatrix = frustum(-0.5, 0.5, -0.5, 0.5, 1.0, 60.0);

	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, &projectionMatrix.m);
	
    glClearColor (1.0, 1.0, 1.0, 1.0);
}


int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_SINGLE | GLUT_RGB | GLUT_DEPTH);
    glutInitContextVersion(3, 2);
    glutInitWindowSize(500, 500);
    glutCreateWindow("GLUGG more shapes");
    glutRepeatingTimer(15);
    myinit();
    glutDisplayFunc(display);
    glutMainLoop();
}
