// This demo is based on glut_shapes.c in the FreeGlut library.
// Original code by Mark J. Kilgard.
// Conversion to a GLUGG demo by Ingemar Ragnemalm 2021.
// Basically shapes converted from glut_shapes pasted in into glutGears.
// Version 2: Even more converted, removing internal (old-style C)
// vector operations in favor of VectorUtils3 and vec3's.
// Also added size for several shapes that didn't have it.

// Copyright statement from original code:

/* Copyright (c) Mark J. Kilgard, 1994, 1997. */

/**
(c) Copyright 1993, Silicon Graphics, Inc.

ALL RIGHTS RESERVED

Permission to use, copy, modify, and distribute this software
for any purpose and without fee is hereby granted, provided
that the above copyright notice appear in all copies and that
both the copyright notice and this permission notice appear in
supporting documentation, and that the name of Silicon
Graphics, Inc. not be used in advertising or publicity
pertaining to distribution of the software without specific,
written prior permission.

THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU
"AS-IS" AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR
OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  IN NO
EVENT SHALL SILICON GRAPHICS, INC.  BE LIABLE TO YOU OR ANYONE
ELSE FOR ANY DIRECT, SPECIAL, INCIDENTAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES WHATSOEVER,
INCLUDING WITHOUT LIMITATION, LOSS OF PROFIT, LOSS OF USE,
SAVINGS OR REVENUE, OR THE CLAIMS OF THIRD PARTIES, WHETHER OR
NOT SILICON GRAPHICS, INC.  HAS BEEN ADVISED OF THE POSSIBILITY
OF SUCH LOSS, HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
ARISING OUT OF OR IN CONNECTION WITH THE POSSESSION, USE OR
PERFORMANCE OF THIS SOFTWARE.

US Government Users Restricted Rights

Use, duplication, or disclosure by the Government is subject to
restrictions set forth in FAR 52.227.19(c)(2) or subparagraph
(c)(1)(ii) of the Rights in Technical Data and Computer
Software clause at DFARS 252.227-7013 and/or in similar or
successor clauses in the FAR or the DOD or NASA FAR
Supplement.  Unpublished-- rights reserved under the copyright
laws of the United States.  Contractor/manufacturer is Silicon
Graphics, Inc., 2011 N.  Shoreline Blvd., Mountain View, CA
94039-7311.

OpenGL(TM) is a trademark of Silicon Graphics, Inc.
*/

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef __APPLE__
	#include <OpenGL/gl3.h>
	#include "MicroGlut.h"
// uses framework Cocoa
#else
	#include <GL/glut.h>
#endif
#include "GL_utilities.h"
#include "VectorUtils3.h"
#include "glugg.h"

#ifndef M_PI
#define M_PI 3.14159265
#endif

static long count = 0;
static long limit = -1;
static float startTime;

static GLfloat view_rotx = 20.0, view_roty = 30.0, view_rotz = 0.0;
static GLuint shape1, shape2, shape3, shape4, shape5, shape6; // Was display lists, now are VAOs
static GLfloat angle = 0.0;

int shape1Triangles=0, shape2Triangles=0, shape3Triangles=0, shape4Triangles=0, shape5Triangles=0, shape6Triangles=0;

GLuint program; // Shader
mat4 projectionMatrix;

static GLfloat lightPos[4] = { 5.0, 5.0, 10.0, 0.0 };

static GLfloat red[4] = { 0.8, 0.1, 0.0, 1.0 };
static GLfloat green[4] = { 0.0, 0.8, 0.2, 1.0 };
static GLfloat blue[4] = { 0.2, 0.2, 1.0, 1.0 };
static GLfloat yellow[4] = { 1, 1, 0.2, 1.0 };
static GLfloat magenta[4] = { 1.0, 0.2, 1.0, 1.0 };
static GLfloat cyan[4] = { 0, 1, 1, 1.0 };


GLuint box(GLfloat size, int *count)
{
  static vec3 n[6] =
  {
    {-1.0, 0.0, 0.0},
    {0.0, 1.0, 0.0},
    {1.0, 0.0, 0.0},
    {0.0, -1.0, 0.0},
    {0.0, 0.0, 1.0},
    {0.0, 0.0, -1.0}
  };
  static GLint faces[6][4] =
  {
    {0, 1, 2, 3},
    {3, 2, 6, 7},
    {7, 6, 5, 4},
    {4, 5, 1, 0},
    {5, 6, 2, 1},
    {7, 4, 0, 3}
  };
  vec3 v[8] =
  {
  	{-size/2,-size/2,-size/2},
  	{-size/2,-size/2, size/2},
  	{-size/2, size/2, size/2},
  	{-size/2, size/2,-size/2},
  	{ size/2,-size/2,-size/2},
  	{ size/2,-size/2, size/2},
  	{ size/2, size/2, size/2},
  	{ size/2, size/2,-size/2}
  };
  GLint i;

  gluggBegin(GLUGG_QUADS);
  
  for (i = 5; i >= 0; i--)
  {
    gluggNormalv(n[i]);
    gluggVertexv(v[faces[i][0]]);
    gluggVertexv(v[faces[i][1]]);
    gluggVertexv(v[faces[i][2]]);
    gluggVertexv(v[faces[i][3]]);
  }
  return gluggEnd(count, program, 0);
}


GLuint doughnut(GLfloat r, GLfloat R, GLint nsides, GLint rings, int *count)
{
  int i, j;
  GLfloat theta, phi, theta1;
  GLfloat cosTheta, sinTheta;
  GLfloat cosTheta1, sinTheta1;
  GLfloat ringDelta, sideDelta;

  ringDelta = 2.0 * M_PI / rings;
  sideDelta = 2.0 * M_PI / nsides;

  theta = 0.0;
  cosTheta = 1.0;
  sinTheta = 0.0;
  gluggBegin(GLUGG_TRIANGLE_STRIP);
  for (i = rings - 1; i >= 0; i--)
  {
    // Reset for each strip.
  	gluggMode(GLUGG_TRIANGLE_STRIP);
  	
    theta1 = theta + ringDelta;
    cosTheta1 = cos(theta1);
    sinTheta1 = sin(theta1);
    
    phi = 0.0;
    for (j = nsides; j >= 0; j--)
    {
      GLfloat cosPhi, sinPhi, dist;
      GLfloat cosPhi2, sinPhi2, dist2;

      phi += sideDelta;
      cosPhi = cos(phi);
      sinPhi = sin(phi);
      dist = R + r * cosPhi;

		// For quads, must have previous edge!
		// TRIANGLE_STRIP does not need this.
//	  GLfloat phi2 = phi - sideDelta;
//	  cosPhi2 = cos(phi2);
//      sinPhi2 = sin(phi2);
//      dist2 = R + r * cosPhi2;

//	x1 = cosTheta1 * cosPhi;
//	y1 = -sinTheta * cosPhi;
//	z1 = r * sinPhi;
// Was GL_QUAD_STRIP
//      gluggNormal(cosTheta1 * cosPhi, -sinTheta1 * cosPhi, sinPhi);
//      gluggVertex(cosTheta1 * dist, -sinTheta1 * dist, r * sinPhi);
//      gluggNormal(cosTheta * cosPhi, -sinTheta * cosPhi, sinPhi);
//      gluggVertex(cosTheta * dist, -sinTheta * dist,  r * sinPhi);

//		vec3 n1 = SetVector(cosTheta1 * cosPhi2, -sinTheta1 * cosPhi2, sinPhi2);
//		vec3 p1 = SetVector(cosTheta1 * dist2, -sinTheta1 * dist2, r * sinPhi2);
		vec3 n2 = SetVector(cosTheta1 * cosPhi, -sinTheta1 * cosPhi, sinPhi);
		vec3 p2 = SetVector(cosTheta1 * dist, -sinTheta1 * dist, r * sinPhi);
		vec3 n3 = SetVector(cosTheta * cosPhi, -sinTheta * cosPhi, sinPhi);
		vec3 p3 = SetVector(cosTheta * dist, -sinTheta * dist,  r * sinPhi);
//		vec3 n4 = SetVector(cosTheta * cosPhi2, -sinTheta * cosPhi2, sinPhi2);
//		vec3 p4 = SetVector(cosTheta * dist2, -sinTheta * dist2, r * sinPhi2);

	// Can be made with GLUGG_QUADS:
	//	gluggNormalv(n4);
	//	gluggVertexv(p4);
	//	gluggNormalv(n3);
	//	gluggVertexv(p3);
	//	gluggNormalv(n2);
	//	gluggVertexv(p2);
	//	gluggNormalv(n1);
	//	gluggVertexv(p1);
	
	// Works with GLUGG_TRIANGLE_STRIP:
		gluggNormalv(n2);
		gluggVertexv(p2);
		gluggNormalv(n3);
		gluggVertexv(p3);
	}
    theta = theta1;
    cosTheta = cosTheta1;
    sinTheta = sinTheta1;
  }
  return gluggEnd(count, program, 0);
}


static vec3 dodec[20];

static void
initDodecahedron(void)
{
  GLfloat alpha, beta;

  alpha = sqrt(2.0 / (3.0 + sqrt(5.0)));
  beta = 1.0 + sqrt(6.0 / (3.0 + sqrt(5.0)) -
    2.0 + 2.0 * sqrt(2.0 / (3.0 + sqrt(5.0))));
  /* *INDENT-OFF* */
  dodec[0].x = -alpha; dodec[0].y = 0; dodec[0].z = beta;
  dodec[1].x = alpha; dodec[1].y = 0; dodec[1].z = beta;
  dodec[2].x = -1; dodec[2].y = -1; dodec[2].z = -1;
  dodec[3].x = -1; dodec[3].y = -1; dodec[3].z = 1;
  dodec[4].x = -1; dodec[4].y = 1; dodec[4].z = -1;
  dodec[5].x = -1; dodec[5].y = 1; dodec[5].z = 1;
  dodec[6].x = 1; dodec[6].y = -1; dodec[6].z = -1;
  dodec[7].x = 1; dodec[7].y = -1; dodec[7].z = 1;
  dodec[8].x = 1; dodec[8].y = 1; dodec[8].z = -1;
  dodec[9].x = 1; dodec[9].y = 1; dodec[9].z = 1;
  dodec[10].x = beta; dodec[10].y = alpha; dodec[10].z = 0;
  dodec[11].x = beta; dodec[11].y = -alpha; dodec[11].z = 0;
  dodec[12].x = -beta; dodec[12].y = alpha; dodec[12].z = 0;
  dodec[13].x = -beta; dodec[13].y = -alpha; dodec[13].z = 0;
  dodec[14].x = -alpha; dodec[14].y = 0; dodec[14].z = -beta;
  dodec[15].x = alpha; dodec[15].y = 0; dodec[15].z = -beta;
  dodec[16].x = 0; dodec[16].y = beta; dodec[16].z = alpha;
  dodec[17].x = 0; dodec[17].y = beta; dodec[17].z = -alpha;
  dodec[18].x = 0; dodec[18].y = -beta; dodec[18].z = alpha;
  dodec[19].x = 0; dodec[19].y = -beta; dodec[19].z = -alpha;
  /* *INDENT-ON* */
}

static void
pentagon(int a, int b, int c, int d, int e, float size)
{
  vec3 n0, d1, d2;

  d1 = VectorSub(dodec[a], dodec[b]);
  d2 = VectorSub(dodec[b], dodec[c]);
  n0 = CrossProduct(d1, d2);
  n0 = normalize(n0);

  gluggNormalv(n0);
  gluggVertexv(ScalarMult(dodec[a], size));
  gluggVertexv(ScalarMult(dodec[b], size));
  gluggVertexv(ScalarMult(dodec[c], size));
  
  gluggVertexv(ScalarMult(dodec[a], size));
  gluggVertexv(ScalarMult(dodec[c], size));
  gluggVertexv(ScalarMult(dodec[d], size));

  gluggVertexv(ScalarMult(dodec[a], size));
  gluggVertexv(ScalarMult(dodec[d], size));
  gluggVertexv(ScalarMult(dodec[e], size));
}

GLuint dodecahedron(float size, int *count)
{
  static int inited = 0;

  if (inited == 0) {
    inited = 1;
    initDodecahedron();
  }
  gluggBegin(GLUGG_TRIANGLES);
  pentagon(0, 1, 9, 16, 5, size);
  pentagon(1, 0, 3, 18, 7, size);
  pentagon(1, 7, 11, 10, 9, size);
  pentagon(11, 7, 18, 19, 6, size);
  pentagon(8, 17, 16, 9, 10, size);
  pentagon(2, 14, 15, 6, 19, size);
  pentagon(2, 13, 12, 4, 14, size);
  pentagon(2, 19, 18, 3, 13, size);
  pentagon(3, 0, 5, 12, 13, size);
  pentagon(6, 15, 8, 10, 11, size);
  pentagon(4, 17, 8, 15, 14, size);
  pentagon(4, 12, 5, 16, 17, size);
  return gluggEnd(count, program, 0);
}

static void
drawtriangle(int i, vec3 data[], int ndx[][3], GLfloat size)
{
  vec3 x0, x1, x2;
  vec3 q0, q1;

  x0 = ScalarMult(data[ndx[i][0]], size);
  x1 = ScalarMult(data[ndx[i][1]], size);
  x2 = ScalarMult(data[ndx[i][2]], size);

  q0 = VectorSub(x0, x1);
  q1 = VectorSub(x1, x2);
  q1 = CrossProduct(q0, q1);
  q1 = normalize(q1);

  gluggNormalv(q1);
  gluggVertexv(x0);
  gluggVertexv(x1);
  gluggVertexv(x2);
}

/* octahedron data: The octahedron produced is centered at the
   origin and has radius 1.0 */
static vec3 odata[6] =
{
  {1.0, 0.0, 0.0},
  {-1.0, 0.0, 0.0},
  {0.0, 1.0, 0.0},
  {0.0, -1.0, 0.0},
  {0.0, 0.0, 1.0},
  {0.0, 0.0, -1.0}
};

static int ondex[8][3] =
{
  {0, 4, 2},
  {1, 2, 4},
  {0, 3, 4},
  {1, 4, 3},
  {0, 2, 5},
  {1, 5, 2},
  {0, 5, 3},
  {1, 3, 5}
};

GLuint octahedron(GLfloat size, int *count)
{
  int i;

  gluggBegin(GLUGG_TRIANGLES);
  for (i = 7; i >= 0; i--)
    drawtriangle(i, odata, ondex, size);
  return gluggEnd(count, program, 0);
}



#define X .525731112119133606
#define Z .850650808352039932

static vec3 idata[12] =
{
  {-X, 0, Z},
  {X, 0, Z},
  {-X, 0, -Z},
  {X, 0, -Z},
  {0, Z, X},
  {0, Z, -X},
  {0, -Z, X},
  {0, -Z, -X},
  {Z, X, 0},
  {-Z, X, 0},
  {Z, -X, 0},
  {-Z, -X, 0}
};

static int iIndex[20][3] =
{
  {0, 4, 1},
  {0, 9, 4},
  {9, 5, 4},
  {4, 5, 8},
  {4, 8, 1},
  {8, 10, 1},
  {8, 3, 10},
  {5, 3, 8},
  {5, 2, 3},
  {2, 7, 3},
  {7, 10, 3},
  {7, 6, 10},
  {7, 11, 6},
  {11, 0, 6},
  {0, 1, 6},
  {6, 1, 10},
  {9, 0, 11},
  {9, 11, 2},
  {9, 2, 5},
  {7, 2, 11},
};

GLuint icosahedron(GLfloat size, int *count)
{
  int i;

  gluggBegin(GLUGG_TRIANGLES);
  for (i = 19; i >= 0; i--)
    drawtriangle(i, idata, iIndex, size);
  return gluggEnd(count, program, 0);
}

/* ENDCENTRY */

/* tetrahedron data: */

#define Ts       1.73205080756887729

static GLfloat tdata[4][3] =
{
  {Ts, Ts, Ts},
  {Ts, -Ts, -Ts},
  {-Ts, Ts, -Ts},
  {-Ts, -Ts, Ts}
};

static int tndex[4][3] =
{
  {0, 1, 3},
  {2, 1, 0},
  {3, 2, 0},
  {1, 2, 3}
};

GLuint tetrahedron(GLfloat size, int *count)
{
  int i;

  gluggBegin(GLUGG_TRIANGLES);
  for (i = 3; i >= 0; i--)
    drawtriangle(i, tdata, tndex, size);
  return gluggEnd(count, program, 0);
}

void draw(void)
{
//	mat4 rx, ry, rz, t, shapeRot;
	mat4 globalR, m;

  angle += 2.0;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
// GLOBAL ROTATION
	globalR = Mult(T(0, 0, -40.0), Mult(Rx(view_rotx*M_PI/180), Mult(Ry(-view_roty*M_PI/180), Rz(view_rotz*M_PI/180))));
	
// shape 1
	m = Mult(globalR, Mult(T(-4.0, -4.0, 0.0), Mult(Mult(Rx(angle/100), Rz(angle*M_PI/180)), S(0.5, 0.5, 0.5))));
	
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, red);
	glBindVertexArray(shape1);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, shape1Triangles);

// shape 2
	m = Mult(globalR, Mult(T(3.1, -3.0, 0.0), Mult(Mult(Rx(angle/100), Rz((-2.0 * angle - 9.0)*M_PI/180)), S(1,1,1))));
	
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, green);
	glBindVertexArray(shape2);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, shape2Triangles);

// shape 3
	m = Mult(globalR, Mult(T(-4.1, 4.2, 0.0), Mult(Mult(Ry(angle/100), Rz((-2.0 * angle - 25.0)*M_PI/180)), S(1,1,1))));
	
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, blue);
	glBindVertexArray(shape3);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, shape3Triangles);
	
// shape 4
	m = Mult(globalR, Mult(T(3.1, 4.2, 0.0), Rz((-2.0 * angle - 25.0)*M_PI/180)));
	
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, yellow);
	glBindVertexArray(shape4);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, shape4Triangles);
	
// shape 5
	m = Mult(globalR, Mult(T(0, 4.2, 0.0), Rz((-2.0 * angle - 25.0)*M_PI/180)));
	
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, magenta);
	glBindVertexArray(shape5);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, shape5Triangles);
	
// shape 6
	m = Mult(globalR, Mult(T(0.1, 0, 0.0), Mult(Ry(angle / 50.0), Rx((-2.0 * angle)*M_PI/180))));
	
	glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, cyan);
	glBindVertexArray(shape6);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, shape6Triangles);
	
	glutSwapBuffers();
}

/* change view angle, exit upon ESC */
/* ARGSUSED1 */
static void
key(unsigned char k, int x, int y)
{
  switch (k)
  {
  case 'z':
    view_rotz += 5.0;
    break;
  case 'Z':
    view_rotz -= 5.0;
    break;
  case 27:  /* Escape */
    exit(0);
    break;
  default:
    return;
  }
  glutPostRedisplay();
}

/* change view angle */
/* ARGSUSED1 */
static void
special(int k, int x, int y)
{
  switch (k)
  {
  case GLUT_KEY_UP:
    view_rotx += 5.0;
    break;
  case GLUT_KEY_DOWN:
    view_rotx -= 5.0;
    break;
  case GLUT_KEY_LEFT:
    view_roty += 5.0;
    break;
  case GLUT_KEY_RIGHT:
    view_roty -= 5.0;
    break;
  default:
    return;
  }
  glutPostRedisplay();
}

/* new window size or exposure */
// This will give us fixed proportions instead of scaling, which is the default.
static void
reshape(int width, int height)
{
	GLfloat h = (GLfloat) height / (GLfloat) width;

	glViewport(0, 0, (GLint) width, (GLint) height);
	projectionMatrix = frustum(-1.0, 1.0, -h, h, 6.0, 60.0);
	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, &projectionMatrix.m);
}


static void
init(void)
{
	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);

	projectionMatrix = frustum(-1.0, 1.0, -1.0, 1.0, 6.0, 60.0);

	program = loadShaders("shapes.vert", "shapes.frag");
	glUniform4fv(glGetUniformLocation(program, "pos"), 1, lightPos);
	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, &projectionMatrix.m);
	
	/* make the shapes */
	shape1 = tetrahedron(2.0, &shape1Triangles);
	shape2 = icosahedron(2.0, &shape2Triangles);
	shape3 = octahedron(2.0, &shape3Triangles);
	shape4 = dodecahedron(1.0, &shape4Triangles);
	shape5 = box(1.5, &shape5Triangles);
	shape6 = doughnut(0.7, 1.5, 20, 20, &shape6Triangles);

	startTime = glutGet(GLUT_ELAPSED_TIME) / 1000.0;
}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);

	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitContextVersion(3, 2);

	glutCreateWindow("GLUGG shapes (glut_shapes version)");
	init();

	glutDisplayFunc(draw);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(key);
	glutSpecialFunc(special);
	glutRepeatingTimer(15);

	glutMainLoop();
	return 0;             /* ANSI C requires main to return int. */
}
