
/* Recursive subdivision of cube (Chapter 6). */

// Source unknown/lost.
// Converted to use GLUGG by Ingemar 2021.
// All that remains of the original is the triangle subdivisiom

#include <stdlib.h>
#ifdef __APPLE__
	#include <OpenGL/gl3.h>
// uses framework Cocoa
#else
	#include <GL/gl.h>
#endif
#include "MicroGlut.h"
#include "GL_utilities.h"
#include "VectorUtils3.h"
#include "glugg.h"

GLuint program; // Shader
mat4 projectionMatrix;

/* initial tetrahedron */

vec3 v[]={{0.0, 0.0, 1.0}, {0.0, 0.942809, -0.33333},
          {-0.816497, -0.471405, -0.333333},
          {0.816497, -0.471405, -0.333333}};

static GLfloat theta[] = {0.0,0.0,0.0};

void triangle( vec3 a, vec3 b, vec3 c)
{
       gluggNormalv(a);
       gluggVertexv(a);
       gluggNormalv(b);
       gluggVertexv(b);
       gluggNormalv(c);
       gluggVertexv(c);
}

void divide_triangle(vec3 a, vec3 b, vec3 c, int m)
{
/* triangle subdivision using vertex numbers
righthand rule applied to create outward pointing faces */

    vec3 v1, v2, v3;
    int j;
    if(m>0)
    {
        v1 = VectorAdd(a, b);
        v1 = normalize(v1);
        v2 = VectorAdd(a, c);
        v2 = normalize(v2);
        v3 = VectorAdd(b, c);
        v3 = normalize(v3);
        divide_triangle(a, v1, v2, m-1);
        divide_triangle(c, v2, v3, m-1);
        divide_triangle(b, v3, v1, m-1);
        divide_triangle(v1, v3, v2, m-1);
    }
    else(triangle(a,b,c)); /* draw triangle at end of recursion */
}

GLuint tetrahedron( int m, int *count)
{
/* Apply triangle subdivision to faces of tetrahedron */
	gluggBegin(GLUGG_TRIANGLES);
    divide_triangle(v[0], v[1], v[2], m);
    divide_triangle(v[3], v[2], v[1], m);
    divide_triangle(v[0], v[3], v[1], m);
    divide_triangle(v[0], v[2], v[3], m);
	return gluggEnd(count, program, 0);
}

static GLfloat red[4] = { 0.8, 0.1, 0.0, 1.0 };
static GLfloat green[4] = { 0.0, 0.8, 0.2, 1.0 };
static GLfloat blue[4] = { 0.2, 0.2, 1.0, 1.0 };
static GLfloat yellow[4] = { 1, 1, 0.2, 1.0 };
static GLfloat magenta[4] = { 1.0, 0.2, 1.0, 1.0 };
static GLfloat cyan[4] = { 0, 1, 1, 1.0 };

GLuint s1, s2, s3;
int count1, count2, count3;

void display(void)
{

/* Displays all three modes, side by side */

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    mat4 cam = lookAt(0, 3, 4, 0, 0, 0, 0,1,0);
	
	GLfloat angle = glutGet(GLUT_ELAPSED_TIME) / 1000.0;
	
    mat4 m = Mult(cam, Mult(Ry(angle), Mult(T(2.0, -1.0,0.0), Rx(angle))));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, red);
	glBindVertexArray(s1);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, count1);

    m = Mult(cam, Mult(Rx(angle), Mult(T(-2.0, -1.0,0.0), Ry(angle))));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, blue);
	glBindVertexArray(s2);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, count2);

    m = Mult(cam, Mult(Mult(Rz(angle), T(0.0, 2.0,0.0)), Ry(angle)));
    glUniformMatrix4fv(glGetUniformLocation(program, "modelviewMatrix"), 1, GL_TRUE, &m.m);
	glUniform4fv(glGetUniformLocation(program, "color"), 1, green);
	glBindVertexArray(s3);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, count3);

    glutSwapBuffers();
}

void myinit()
{
    glEnable(GL_DEPTH_TEST); /* enable z buffer */
    glDisable(GL_CULL_FACE);

	program = loadShaders("sphere.vert", "sphere.frag");

	s1 = tetrahedron(1, &count1);
	s2 = tetrahedron(3, &count2);
	s3 = tetrahedron(4, &count3);

	projectionMatrix = frustum(-1.0, 1.0, -1.0, 1.0, 1.0, 60.0);

	glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, &projectionMatrix.m);

    glClearColor (1.0, 1.0, 1.0, 1.0);
    glColor3f (0.0, 0.0, 0.0);
}


int main(int argc, char **argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_SINGLE | GLUT_RGB | GLUT_DEPTH);
    glutInitContextVersion(3, 2);
    glutInitWindowSize(500, 500);
    glutRepeatingTimer(15);
    glutCreateWindow("GLUGG sphere");
    myinit();
    glutDisplayFunc(display);
    glutMainLoop();
}
