// Demo of the different recording modes, an extension of the gluggTriangle demos.
//( By Ingemar 2021.

#ifdef __APPLE__
	#include <OpenGL/gl3.h>
	#include "MicroGlut.h"
// uses framework Cocoa
#else
	#include <GL/glut.h>
#endif
#include "GL_utilities.h"
#include "VectorUtils3.h"
#include "glugg.h"

GLuint program; // Shader

int trianglesCount, quadCount, fanCount, stripCount;
GLuint triangles, quads, fan, strip;

void draw(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_CULL_FACE);
	
	glBindVertexArray(triangles);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, trianglesCount);

	glBindVertexArray(quads);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, quadCount);

	glBindVertexArray(fan);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, fanCount);

	glBindVertexArray(strip);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, stripCount);

	glutSwapBuffers();
}


void init(void)
{
	program = loadShaders("color.vert", "color.frag");
	
	// For this demo, I make one shape (buffer) in each mode.
	// Note that it is legal to switch mode withon one shape.
	
	glDisable(GL_CULL_FACE);
	
	/* make some triangles */
	gluggBegin(GLUGG_TRIANGLES);

	gluggNormal(1, 0, 0); // Red
	
	gluggVertex(0.1, 0.6,0);
	gluggVertex(0.4, 0.7,0);
	gluggVertex(0.3, 0.8,0);

	gluggVertex(0.6, 0.9,0);
	gluggVertex(0.7, 0.7,0);
	gluggVertex(0.9, 0.8,0);
	
	gluggVertex(0.5, 0.5,0);
	gluggVertex(0.4, 0.2,0);
	gluggVertex(0.7, 0.3,0);
	
	triangles = gluggEnd(&trianglesCount, program, 0);

	// Make some quads
	gluggBegin(GLUGG_QUADS);

	gluggNormal(0, 1, 0); // Green

	gluggVertex(0.2, -0.5, 0);
	gluggVertex(0.4, -0.4, 0);
	gluggVertex(0.3, -0.2, 0);
	gluggVertex(0.1, -0.3, 0);

	gluggVertex(0.5, -0.9, 0);
	gluggVertex(0.6, -0.7, 0);
	gluggVertex(0.4, -0.6, 0);
	gluggVertex(0.3, -0.8, 0);

	gluggVertex(0.5, -0.5, 0);
	gluggVertex(0.7, -0.5, 0);
	gluggVertex(0.7, -0.3, 0);
	gluggVertex(0.5, -0.3, 0);

	quads = gluggEnd(&quadCount, program, 0);

	// Make a fan
	gluggBegin(GLUGG_TRIANGLE_FAN);

	gluggNormal(0, 1, 1); // Cyan

	gluggVertex(-0.5, -0.5, 0); // Center

	gluggVertex(-0.7, -0.3, 0);
	gluggVertex(-0.8, -0.5, 0);
	gluggVertex(-0.7, -0.7, 0);
	gluggVertex(-0.5, -0.8, 0);
	gluggVertex(-0.3, -0.7, 0);
	gluggVertex(-0.2, -0.5, 0);
	gluggVertex(-0.3, -0.3, 0);
	gluggVertex(-0.5, -0.2, 0); // Intentionally missing one vertex!

	fan = gluggEnd(&fanCount, program, 0);

	// Make a strip
	gluggBegin(GLUGG_TRIANGLE_STRIP);

	gluggNormal(1, 1, 0); // Yellow

	gluggVertex(-0.3, 0.1, 0);
	gluggVertex(-0.2, 0.3, 0);
	gluggVertex(-0.5, 0.3, 0);
	gluggVertex(-0.3, 0.4, 0);
	gluggVertex(-0.5, 0.5, 0);
	gluggVertex(-0.3, 0.6, 0);
	gluggVertex(-0.7, 0.7, 0);
	gluggVertex(-0.5, 0.8, 0);

	strip = gluggEnd(&stripCount, program, 0);

}

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutInitContextVersion(3, 2);

	glutCreateWindow("GLUGG Modes");
	init();
	glutDisplayFunc(draw);

	glutMainLoop();
	return 0;
}
